@@ -1,5 +1,6 @@
 mod args;
 mod network;
+mod time;
 mod utils;
 mod vendor;
 
@@ -10,9 +11,7 @@ use std::sync::Arc;
 use std::time::Duration;
 use std::sync::atomic::{AtomicBool, Ordering};
 
-use ipnetwork::NetworkSize;
 use pnet::datalink;
-use rand::prelude::*;
 
 use crate::args::{ScanOptions, OutputFormat};
 use crate::vendor::Vendor;
@@ -47,35 +46,10 @@ fn main() {
         process::exit(1);
     }
 
-    let selected_interface: &datalink::NetworkInterface = interfaces.iter()
-        .find(|interface| { interface.name == scan_options.interface_name && interface.is_up() && !interface.is_loopback() })
-        .unwrap_or_else(|| {
-            eprintln!("Could not find interface with name {}", scan_options.interface_name);
-            process::exit(1);
-        });
-
-    let ip_network = match selected_interface.ips.first() {
-        Some(ip_network) if ip_network.is_ipv4() => ip_network,
-        Some(_) => {
-            eprintln!("Only IPv4 networks supported");
-            process::exit(1);
-        },
-        None => {
-            eprintln!("Expects a valid IP on the interface, none found");
-            process::exit(1);
-        }
-    };
+    let (selected_interface, ip_networks) = network::compute_network_configuration(&interfaces, &scan_options);
 
     if scan_options.is_plain_output() {
-
-        println!();
-        println!("Selected interface {} with IP {}", selected_interface.name, ip_network);
-        if let Some(forced_source_ipv4) = scan_options.source_ipv4 {
-            println!("The ARP source IPv4 will be forced to {}", forced_source_ipv4);
-        }
-        if let Some(forced_destination_mac) = scan_options.destination_mac {
-            println!("The ARP destination MAC will be forced to {}", forced_destination_mac);
-        }
+        utils::display_prescan_details(&ip_networks, selected_interface, scan_options.clone());
     }
 
     // Start ARP scan operation
@@ -85,7 +59,7 @@ fn main() {
     // local network.
 
     let channel_config = datalink::Config {
-        read_timeout: Some(Duration::from_millis(500)), 
+        read_timeout: Some(Duration::from_millis(network::DATALINK_RCV_TIMEOUT)), 
         ..datalink::Config::default()
     };
 
@@ -112,41 +86,53 @@ fn main() {
     let cloned_options = Arc::clone(&scan_options);
     let arp_responses = thread::spawn(move || network::receive_arp_responses(&mut rx, cloned_options, cloned_timed_out, &mut vendor_list));
 
-    let network_size: u128 = match ip_network.size() {
-        NetworkSize::V4(ipv4_network_size) => ipv4_network_size.into(),
-        NetworkSize::V6(_) => {
-            eprintln!("IPv6 networks are not supported by the ARP protocol");
-            process::exit(1);
-        }
-    };
+    let network_size = utils::compute_network_size(&ip_networks);
+
+    let estimations = network::compute_scan_estimation(network_size, &scan_options);
+    let interval_ms = estimations.interval_ms;
 
     if scan_options.is_plain_output() {
-        println!("Sending {} ARP requests (waiting at least {}ms, {}ms request interval)", network_size, scan_options.timeout_ms, scan_options.interval_ms);
+
+        let formatted_ms = time::format_milliseconds(estimations.duration_ms);
+        println!("Estimated scan time {} ({} bytes, {} bytes/s)", formatted_ms, estimations.request_size, estimations.bandwidth);
+        println!("Sending {} ARP requests (waiting at least {}ms, {}ms request interval)", network_size, scan_options.timeout_ms, interval_ms);
     }
 
+    let finish_sleep = Arc::new(AtomicBool::new(false));
+    let cloned_finish_sleep = Arc::clone(&finish_sleep);
+
+    ctrlc::set_handler(move || {
+        eprintln!("[warn] Receiving halt signal, ending scan with partial results");
+        cloned_finish_sleep.store(true, Ordering::Relaxed);
+    }).unwrap_or_else(|err| {
+        eprintln!("Could not set CTRL+C handler ({})", err);
+        process::exit(1);
+    });
+
     // The retry count does right now use a 'brute-force' strategy without
     // synchronization process with the already known hosts.
     for _ in 0..scan_options.retry_count {
 
+        if finish_sleep.load(Ordering::Relaxed) {
+            break;
+        }
+
         // The random approach has one major drawback, compared with the native
         // network iterator exposed by 'ipnetwork': memory usage. Instead of
         // using a small memory footprint iterator, we have to store all IP
         // addresses in memory at once. This can cause problems on large ranges.
-        let ip_addresses: Vec<IpAddr> = match scan_options.randomize_targets {
-            true => {
-                let mut rng = rand::thread_rng();
-                let mut shuffled_addresses: Vec<IpAddr> = ip_network.iter().collect();
-                shuffled_addresses.shuffle(&mut rng);
-                shuffled_addresses
-            },
-            false => ip_network.iter().collect()
-        };
+        let ip_addresses: Vec<IpAddr> = network::compute_ip_range(&ip_networks, &scan_options);
+        let source_ip = network::find_source_ip(selected_interface, scan_options.source_ipv4);
 
         for ip_address in ip_addresses {
 
+            if finish_sleep.load(Ordering::Relaxed) {
+                break;
+            }
+
             if let IpAddr::V4(ipv4_address) = ip_address {
-                network::send_arp_request(&mut tx, selected_interface, &ip_network, ipv4_address, Arc::clone(&scan_options));
-                thread::sleep(Duration::from_millis(scan_options.interval_ms));
+                network::send_arp_request(&mut tx, selected_interface, source_ip, ipv4_address, Arc::clone(&scan_options));
+                thread::sleep(Duration::from_millis(interval_ms));
             }
         }
     }
@@ -154,7 +140,12 @@ fn main() {
     // Once the ARP packets are sent, the main thread will sleep for T seconds
     // (where T is the timeout option). After the sleep phase, the response
     // thread will receive a stop request through the 'timed_out' mutex.
-    thread::sleep(Duration::from_millis(scan_options.timeout_ms));
+    let mut sleep_ms_mount: u64 = 0;
+    while !finish_sleep.load(Ordering::Relaxed) && sleep_ms_mount < scan_options.timeout_ms {
+        
+        thread::sleep(Duration::from_millis(500));
+        sleep_ms_mount += 500;
+    }
     timed_out.store(true, Ordering::Relaxed);
 
     let (response_summary, target_details) = arp_responses.join().unwrap_or_else(|error| {
@@ -165,6 +156,7 @@ fn main() {
     match &scan_options.output {
         OutputFormat::Plain => utils::display_scan_results(response_summary, target_details, &scan_options),
         OutputFormat::Json => println!("{}", utils::export_to_json(response_summary, target_details)),
-        OutputFormat::Yaml => println!("{}", utils::export_to_yaml(response_summary, target_details))
+        OutputFormat::Yaml => println!("{}", utils::export_to_yaml(response_summary, target_details)),
+        OutputFormat::Csv => print!("{}", utils::export_to_csv(response_summary, target_details))
     }
 }
